<?php

namespace Drupal\melhor_envio\Plugin\Commerce\ShippingMethod;

use Drupal\commerce_price\Price;
use Drupal\commerce_shipping\Entity\ShipmentInterface;
use Drupal\commerce_shipping\PackageTypeManagerInterface;
use Drupal\commerce_shipping\Plugin\Commerce\ShippingMethod\ShippingMethodBase;
use Drupal\commerce_shipping\ShippingRate;
use Drupal\Core\Form\FormStateInterface;
use Drupal\state_machine\WorkflowManagerInterface;
use MelhorEnvio\Enums\Environment;
use MelhorEnvio\Resources\Shipment\Product;
use MelhorEnvio\Shipment as MelhorEnvioShipment;

/**
 * Provides the Melhor Envio shipping method.
 *
 * @CommerceShippingMethod(
 *   id = "melhor_envio",
 *   label = @Translation("Melhor Envio"),
 *   services = {
 *     1 = @Translation("PAC Correios"),
 *     2 = @Translation("SEDEX Correios"),
 *     14 = @Translation("ESEDEX Correios"),
 *     17 = @Translation("MINI Correios"),
 *     3 = @Translation("JADLOG Pacote"),
 *     4 = @Translation("JADLOG COM"),
 *     8 = @Translation("VIABRASIL Aéreo"),
 *     9 = @Translation("VIABRASIL Rodoviario"),
 *     10 = @Translation("Próximo dia LATAM"),
 *     11 = @Translation("Próximo Voo LATAM"),
 *     12 = @Translation("Juntos LATAM"),
 *     15 = @Translation("Amanhã Azul"),
 *     16 = @Translation("E-commerce AZUL"),
 *   }
 * )
 */
class MelhorEnvio extends ShippingMethodBase {

  /**
   * Constructs a new FlatRate object.
   *
   * @param array $configuration
   *   A configuration array containing information about the plugin instance.
   * @param string $plugin_id
   *   The plugin_id for the plugin instance.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   * @param \Drupal\commerce_shipping\PackageTypeManagerInterface $package_type_manager
   *   The package type manager.
   * @param \Drupal\state_machine\WorkflowManagerInterface $workflow_manager
   *   The workflow manager.
   */
  public function __construct(array $configuration, $plugin_id, $plugin_definition, PackageTypeManagerInterface $package_type_manager, WorkflowManagerInterface $workflow_manager) {
    parent::__construct($configuration, $plugin_id, $plugin_definition, $package_type_manager, $workflow_manager);
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    return [
      'rate_label' => '',
      'rate_description' => '',
      'rate_token' => '',
      'rate_amount' => NULL,
      'rate_environment' => '',
      'rate_cep_origin' => '',
    ] + parent::defaultConfiguration();
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {
    $form = parent::buildConfigurationForm($form, $form_state);

    $amount = $this->configuration['rate_amount'];
    // A bug in the plugin_select form element causes $amount to be incomplete.
    if (isset($amount) && !isset($amount['number'], $amount['currency_code'])) {
      $amount = NULL;
    }

    $form['rate_label'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Rate label'),
      '#description' => $this->t('Shown to customers when selecting the rate.'),
      '#default_value' => $this->configuration['rate_label'],
      '#required' => TRUE,
    ];

    $form['rate_environment'] = [
      '#type' => 'radios',
      '#title' => $this->t('Rate environment'),
      '#description' => $this->t('The enviroment.'),
      '#options' => [Environment::SANDBOX => 'Sandbox (test)', Environment::PRODUCTION => 'Production'],
      '#default_value' => $this->configuration['rate_environment'],
    ];

    $form['rate_cep_origin'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Rate CEP origin'),
      '#description' => $this->t('The origin postal code(Only numbers).'),
      '#default_value' => $this->configuration['rate_cep_origin'],
    ];

    $form['rate_description'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Rate description'),
      '#description' => $this->t('Provides additional details about the rate to the customer.'),
      '#default_value' => $this->configuration['rate_description'],
    ];

    $form['rate_token'] = [
      '#type' => 'textarea',
      '#title' => $this->t('Token Melhor Envio'),
      '#description' => $this->t('Provides additional details about the rate to the customer.'),
      '#default_value' => $this->configuration['rate_token'],
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state) {
    parent::submitConfigurationForm($form, $form_state);

    if (!$form_state->getErrors()) {
      $values = $form_state->getValue($form['#parents']);

      $this->configuration['rate_label'] = $values['rate_label'];
      $this->configuration['rate_description'] = $values['rate_description'];
      $this->configuration['rate_cep_origin'] = $values['rate_cep_origin'];
      $this->configuration['rate_environment'] = $values['rate_environment'];
      $this->configuration['rate_token'] = $values['rate_token'];
    }
  }

  public function convertToCentimeters(string $unitMeasurement, float $value) : float {
    return match ($unitMeasurement) {
      'mm' => $value / 10,
      'm' => (100 * $value),
      'km' => (100000 * $value),
      'in' => (2.54 * $value),
      'ft' => (30.48 * $value),
      'M' => ((1.852 * 100) * $value),
      default => $value,
    };
  }

  public function convertToKilograms(string $unitMeasurement, float $value) {
    return match ($unitMeasurement) {
      'g' => $value / 1000,
      'mg' => $value / 1000000,
      'oz' => $value / 35.274,
      'lb' => $value / 2.205,
      'ct' => $value / 5000,
      default => $value,
    };
  }

  public function clearPostalCode(string $postalCode): string {
    return str_replace(['-', '_', ' '], ['', '',''], $postalCode);
  }


  /**
   * Calculates rates for the given shipment.
   *
   * @param \Drupal\commerce_shipping\Entity\ShipmentInterface $shipment
   *   The shipment.
   */
  public function getQuotes(ShipmentInterface $shipment) {
    /** @var \Drupal\address\Plugin\Field\FieldType\AddressItem $address */
    $address = $shipment->getShippingProfile()->get('address')->first();
    $postalCodeDestiny = $this->clearPostalCode($address->getPostalCode());

    $melhorEnvioShipment = new MelhorEnvioShipment($this->configuration['rate_token'], $this->configuration['rate_environment']);
    $calculator = $melhorEnvioShipment->calculator();
    $calculator->postalCode($this->configuration['rate_cep_origin'], $postalCodeDestiny);
    $order = $shipment->getOrder();

    /** @var \Drupal\commerce_order\Entity\OrderItem $item */
    foreach ($order->getItems() as $item) {
      $dimensions = $item->getPurchasedEntity()->get('dimensions')->first();
      $objWeight = $item->getPurchasedEntity()->get('weight')->first();
      $height = $this->convertToCentimeters($dimensions->unit, $dimensions->height);
      $width = $this->convertToCentimeters($dimensions->unit, $dimensions->width);
      $length = $this->convertToCentimeters($dimensions->unit, $dimensions->length);
      $weight = $this->convertToKilograms($objWeight->unit, $objWeight->number);
      $price = $item->getPurchasedEntity()->get('price')->number;

      $calculator->addProduct(new Product(uniqid(), $height, $width, $length, $weight, $price, $item->getQuantity()));
    }

    $calculator->addServices(
      array_keys($this->getServices())
    );

    return $calculator->calculate();
  }

  /**
   * {@inheritdoc}
   */
  public function calculateRates(ShipmentInterface $shipment) {

    $rates = [];
    $quotes = $this->getQuotes($shipment);

    foreach ($quotes as $key => $quote) {
      if (isset($quote['error'])) {
        continue;
      }

      $rates[] = new ShippingRate([
        'shipping_method_id' => $this->parentEntity->id(),
        'service' => $this->services[$quote['id']],
        'amount' => Price::fromArray([
          "number" => $quote['price'],
          "currency_code" => "BRL",
        ]),
        'description' => "De {$quote['custom_delivery_range']['min']} a {$quote['custom_delivery_range']['max']} dias uteis.",
      ]);
    }

    return $rates;
  }

}
